# -*- coding: utf-8 -*-
"""
Print to stdout command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, parser

from safe.app import safe
from safe.ec import NO_SUCH_ACCOUNT, NO_SUCH_QUESTION, PASSWORD_NOT_SET
from safe.model import Account, Password


@safe(name='print')
def print_():
    """Print secret to stdout."""
    parser.add_argument(
        'name',
        nargs=1,
        help='name of the account for which to print secret',
    )
    parser.add_argument(
        '-q',
        '--question',
        help='print answer to security question with given identifier to '
             'clipboard instead of the account password',
    )

    yield

    account = Account.for_slug(args.name[0])
    if account is None:
        print('error: no account named', args.name[0], file=sys.stderr)
        yield NO_SUCH_ACCOUNT

    if args.question:
        question = account.question_query\
                          .filter_by(identifier=args.question)\
                          .first()
        if question is None:
            fmt = 'error: no question with identifier "%s" associated with ' \
                  'account "%s"'
            print(fmt % (args.question, account.name), file=sys.stderr)
            yield NO_SUCH_QUESTION
        value = question.answer
    else:
        password = account.password_query\
                          .order_by(Password.changed.desc())\
                          .limit(1)\
                          .first()
        if password is None:
            msg = 'error: no password set for account "%s"' % account.name
            print(msg, file=sys.stderr)
            yield PASSWORD_NOT_SET
        value = password.value

    print(value)



# -*- coding: utf-8 -*-
"""
Update account command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
from __future__ import print_function

import getpass
import sys

from clik import args, g, parser

from safe.cmd.update import update
from safe.compat import input
from safe.ec import NO_SUCH_ACCOUNT, VALIDATION_ERROR
from safe.form.account import Operation, UpdateAccountForm
from safe.model import Account, Password


@update(alias='a')
def account():
    """Update an account and/or its associated data."""
    parser.add_argument(
        'account',
        help='name or alias of account to update',
        nargs=1,
    )
    parser.add_argument(
        '-p',
        '--password',
        action='store_true',
        default=False,
        help='set the password for an account (prompts for value)',
    )

    form = UpdateAccountForm()
    form.configure_parser()

    yield

    account = Account.for_slug(args.account[0])
    if account is None:
        print('error: no account with name/alias:', args.account)
        yield NO_SUCH_ACCOUNT

    if not form.bind_and_validate(account):
        msg = 'error: there were validation error(s) with input value(s)'
        print(msg, file=sys.stderr)
        form.print_errors()
        yield VALIDATION_ERROR

    new, update = {}, {}
    operations = form.question.operations
    for i, (op, subject, details) in enumerate(operations):
        if op == Operation.Q and not details:
            update.setdefault(subject.identifier, [subject, None, None])
            update[subject.identifier][1] = i
        elif op == Operation.A and not details:
            update.setdefault(subject.identifier, [subject, None, None])
            update[subject.identifier][2] = i
        elif op == Operation.NEW:
            new[subject] = i

    for identifier in sorted(new):
        print('\nAdding new security question with identifier', identifier)
        operations[new[identifier]][2] = input('Question: '), input('Answer: ')

    for identifier in sorted(update):
        print('\nUpdating security question with identifier', identifier)
        question, question_i, answer_i = update[identifier]
        for label, attr in (('Question', 'question'), ('Answer', 'answer')):
            value = getattr(question, attr)
            if not value:
                value = '<empty>'
            print('%s: %s' % (label, value))
        if question_i is not None:
            operations[question_i][2] = input('New question: ')
        if answer_i is not None:
            operations[answer_i][2] = input('New answer: ')

    if args.password:
        print('\nUpdating password for account')
        while True:
            password = getpass.getpass('New password: ')
            confirm = getpass.getpass('Confirm: ')
            if password == confirm:
                break
            print('error: passwords did not match\n', file=sys.stderr)
        g.db.add(Password(account_id=account.id, value=password))

    form.update_account()
    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Update object command group.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
from safe.app import safe


@safe(alias='up')
def update():
    """Update objects in the database."""
    yield



# -*- coding: utf-8 -*-
"""
Shell command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
import subprocess

from clik_shell import DefaultShell, exclude_from_shell

from safe.app import safe
from safe.util import get_executable


CLEAR_EXECUTABLE = get_executable('clear')


class Shell(DefaultShell):
    """Shell subclass that adds the ``clear`` command if available."""

    intro = '\nWelcome to the safe shell. Enter ? for a list of commands.\n'

    def __init__(self):
        """Instantiate the command shell."""
        super(Shell, self).__init__(safe)

    if CLEAR_EXECUTABLE is not None:
        def do_clear(self, _):
            """Clear the screen."""
            subprocess.call(CLEAR_EXECUTABLE)


@exclude_from_shell
@safe(alias='sh')
def shell():
    """Interactive command shell for safe."""
    yield
    Shell().cmdloop()



# -*- coding: utf-8 -*-
"""
IPython shell.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
from safe.app import safe

try:
    import IPython
    ipython_available = True
except ImportError:
    ipython_available = False


ENABLE_IPYTHON = False


if ENABLE_IPYTHON and ipython_available:
    @safe
    def ipy():
        """Open an IPython shell."""
        yield

        from clik import args, g  # noqa: F401
        from safe.model import Account, Alias, Code, Password  # noqa: F401
        from safe.model import Policy, Question  # noqa: F401

        IPython.embed()



# -*- coding: utf-8 -*-
"""
Form definitions.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
from wtforms.validators import ValidationError

from safe.model import SLUG_RE, SLUG_VALIDATION_ERROR_MESSAGE


def slug_validator(_, field):
    """
    Validate that ``field.data`` is a slug.

    :param field: Field to validate
    :type field: :class:`wtforms.fields.core.Field`
    :raises: :exc:`wtforms.validators.ValidationError` if data is invalid
    """
    if field.data and not SLUG_RE.search(field.data):
        raise ValidationError(SLUG_VALIDATION_ERROR_MESSAGE)



# -*- coding: utf-8 -*-
"""
Output utilities for the list command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2019.
:license: BSD
"""
NONE = '<none>'
NOT_SET = '<not set>'


def nullable(value):
    """Return :data:`NOT_SET` value if ``value`` is ``None``, else value."""
    if value is not None:
        return value
    return NOT_SET


def get_widths(rows, start_counts=None):
    """
    Return an list of integers of widest values for each column in ``rows``.

    If ``start_counts`` is not given, the return value is initialized to all
    zeroes.

    :param rows: Sequence of sequences representing rows
    :param list start_counts: Optional start counts for each column
    """
    rv = start_counts
    if rv is None:
        rv = [0] * len(rows[0])
    for row in rows:
        for i, cell in enumerate(row):
            n = len(str(cell))
            if n > rv[i]:
                rv[i] = n
    return rv


def print_table(headers, rows):
    """
    Print a table with ``headers`` and ``rows``.

    :param headers: Sequence of strings to be printed at the top of the table
    :param rows: Sequence of sequences representing the rows
    """
    widths = get_widths(rows, start_counts=[len(h) for h in headers])
    print(('    '.join(['%%-%is' % n for n in widths])) % headers)
    per = len(headers)
    fmt = '── '.join(['%s %s' for _ in range(per)]).rstrip('%s').rstrip()
    for row in rows:
        args = []
        for i, cell in enumerate(row):
            args.extend((cell, '─' * (widths[i] - len(str(cell)))))
        args.pop()
        print(fmt % tuple(args))


def print_detail(title, items, hack=None):
    """
    Print 'detail table' with ``title`` for ``items``.

    :param str title: Printed left-aligned
    :param items: Sequence of 2-tuples ``(field name, value)``, both of
                  which are strings
    :param hack: Ignore this dumb parameter
    """
    print(title)
    fmt = '  ┠%s┄┈ %s → %s'
    xmt = '  ┖%s┄┈ %s → %s'
    x = len(items) - 1
    width = get_widths(items)[0]
    for i, (key, value) in enumerate(items):
        bars = '─' * (width - len(key))
        if i == x:
            if hack:
                index, replacement = hack
                bars = bars[:index] + replacement + bars[index + 1:]
            print(xmt % (bars, key, value))
        else:
            print(fmt % (bars, key, value))


def sorted_by_name(query):
    """Return ``query`` sorted case-insensitively by object ``name``."""
    return sorted(query.all(), key=lambda obj: obj.name.lower())
